@@ -2,6 +2,7 @@ import math
 import re
 import pyproj
 import warnings
+import abc
 
 '''
 Methods for representing geographic coordinates (latitude and longitude)
@@ -15,11 +16,15 @@ Features:
 Written July 22, 2014
 Author: Gen Del Raye
 '''
+# TODO: Write methods to convert -180 to 180 longitudes to 0 to 360 and vice versa
 
 class GeoCoord:
     '''
-    Object representing geographic coordinates (i.e. latitude or longitude)
+    Abstract class representing geographic coordinates (i.e. latitude or longitude)
+    Not meant to be used directly - access through Subclasses Latitude() and Longitude() 
     '''
+    __metaclass__ = abc.ABCMeta
+    
     def __init__(self, degree = 0, minute = 0, second = 0):
         '''
         Initialize a GeoCoord object
@@ -43,6 +48,30 @@ class GeoCoord:
     
     def set_degree(self, degree):
         self.degree = float(degree)
+    
+    @staticmethod
+    def _calc_decimaldegree(degree, minute, second):
+        '''
+        Calculate decimal degree form degree, minute, second
+        '''
+        return degree + minute/60. + second/3600.
+    
+    @staticmethod
+    def _calc_degreeminutes(decimal_degree):
+        '''
+        Calculate degree, minute second from decimal degree
+        '''
+        sign = cmp(decimal_degree, 0) # Store whether the coordinate is negative or positive
+        decimal_degree = abs(decimal_degree)
+        degree = decimal_degree//1 # Truncate degree to be an integer
+        decimal_minute = (decimal_degree - degree)*60. # Calculate the decimal minutes
+        minute = decimal_minute//1 # Truncate minute to be an integer
+        second = (decimal_minute - minute)*60. # Calculate the decimal seconds
+        # Finally, re-impose the appropriate sign
+        degree = degree*sign
+        minute = minute*sign
+        second = second*sign
+        return (degree, minute, decimal_minute, second)
             
     def _update(self):
         '''
@@ -50,24 +79,17 @@ class GeoCoord:
         consistent (for example, if minutes > 60, add extra to degrees, or if degrees is
         a decimal, add extra to minutes).
         '''
-        self.decimal_degree = self.degree + self.minute/60. + self.second/3600.
-        sign = cmp(self.decimal_degree, 0) # Store whether the coordinate is negative or positive
-        decimal_degree = abs(self.decimal_degree)
-        self.degree = decimal_degree//1 # Truncate degree to be an integer
-        self.decimal_minute = (decimal_degree - self.degree)*60. # Calculate the decimal minutes
-        self.minute = self.decimal_minute//1 # Truncate minute to be an integer
-        self.second = (self.decimal_minute - self.minute)*60. # Calculate the decimal seconds
-        # Finally, re-impose the appropriate sign
-        self.degree = self.degree*sign
-        self.minute = self.minute*sign
-        self.second = self.second*sign
+        self.decimal_degree = self._calc_decimaldegree(self.degree, self.minute, self.second)
+        self.degree, self.minute, self.decimal_minute, self.second = self._calc_degreeminutes(self.decimal_degree)
     
+    @abc.abstractmethod
     def get_hemisphere(self):
         '''
         Dummy method, used in child classes such as Latitude and Longitude
         '''
         pass
     
+    @abc.abstractmethod
     def set_hemisphere(self):
         '''
         Dummy method, used in child classes such as Latitude and Longitude
@@ -201,7 +223,29 @@ class Latitude(GeoCoord):
 class Longitude(GeoCoord):
     '''
     Coordinate object specific for longitude coordinates
+    Langitudes outside the range -180 to 180 (i.e. those reported in the range 0 to 360) will
+    automatically be converted to 0 to 360 to ensure that all operations such as hemisphere
+    assignment work as expected. To report in the range 0 to 360, use method range360()
     '''
+    def __init__(self, degree = 0, minute = 0, second = 0):
+        super(Longitude, self).__init__(degree, minute, second) # Initialize the GeoCoord
+        decimal_degree = self.range180() # Make sure that longitudes are reported in the range -180 to 180
+        self.degree, self.minute, self.decimal_minute, self.second = self._calc_degreeminutes(decimal_degree)
+        self._update()
+    
+    def range180(self):
+        '''
+        Report longitudes using the range -180 to 180.
+        '''
+        return ((self.decimal_degree + 180)%360) - 180
+        
+        
+    def range360(self):
+        '''
+        Report longitudes using the range 0 to 360
+        '''
+        return (self.decimal_degree + 360)%360
+        
     def get_hemisphere(self):
         '''
         Returns the hemisphere identifier for the current coordinate
@@ -316,7 +360,7 @@ class LatLon:
         Projection should be a basemap or pyproj projection object or similar
         '''
         x, y = projection(self.lon.decimal_degree, self.lat.decimal_degree)
-        return (y, x)
+        return (x, y)
     
     def complex(self):
         '''
@@ -335,6 +379,8 @@ class LatLon:
         g = pyproj.Geod(ellps = ellipse)
         heading_initial, heading_reverse, distance = g.inv(lon1, lat1, lon2, lat2, radians = False)
         distance = distance/1000.0
+        if heading_initial == 0.0: # Reverse heading not well handled for coordinates that are directly south
+            heading_reverse = 180.0
         return {'heading_initial': heading_initial, 'heading_reverse': heading_reverse, 'distance': distance}
         
     def heading_initial(self, other, **kwargs):
@@ -420,6 +466,26 @@ class LatLon:
         heading = inv['heading_reverse']
         distance = inv['distance']
         return GeoVector(initial_heading = heading, distance = distance)
+    
+    def almost_equal(self, other, e = 0.000001):
+        '''
+        Sometimes required for comparing LatLon coordinates if float error has
+        occurred. Determine if self and other (another LatLon coordinate) are
+        equal to within e km of each other. The default (e = 0.000001) will return
+        True if self and other are less than 1 mm apart in distance.
+        '''
+        return (self - other).magnitude < e
+    
+    def __eq__(self, other):
+        # other is a LatLon object
+        if self.lat == other.lat and self.lon == other.lon:
+            return True
+        else:
+            return False
+        
+    def __ne__(self, other):
+        # other is a LatLon object
+        return not self.__eq__(other)
         
     def __add__(self, other):
         # other is a GeoVector object
@@ -500,8 +566,9 @@ class GeoVector:
             distance (scalar) - the magnitude of the vector in km
         '''
         if dx == None and dy == None: # If only initial_heading and distance are given
-            self.theta = initial_heading
-            theta_rad = math.radians(self.theta)
+            self.heading = initial_heading
+            theta = self._angle_or_heading(self.heading) # Convert heading to angle
+            theta_rad = math.radians(theta)
             self.magnitude = distance
             self.dx = self.magnitude * math.cos(theta_rad)
             self.dy = self.magnitude * math.sin(theta_rad)
@@ -513,14 +580,42 @@ class GeoVector:
             raise NameError, 'Class GeoVector requires two arguments (dx and dy or initial_heading and distance)'
         
     def __call__(self):
-        return self.theta, self.magnitude
+        return self.heading, self.magnitude
+    
+    def _angle_or_heading(self, angle_or_heading):
+        '''
+        Convert angle degrees (i.e. starting at coordinates (1, 0) or 
+        due East and going clockwise to 360) into heading (i.e. starting 
+        at coordinates (0, 1) or due North and going counterclockwise to 
+        360) or vice versa.
+        '''
+        heading_or_angle = (90 - angle_or_heading)%360
+        return heading_or_angle
     
     def _update(self):
         '''
         Calculate heading and distance from dx and dy
         '''
-        self.theta = math.degrees(math.atan(float(self.dy)/self.dx))
-        self.magnitude = math.cos(self.theta) * self.dx
+        try:
+            theta_radians = math.atan(float(self.dy)/self.dx)
+        except ZeroDivisionError:
+            if self.dy > 0: theta_radians = 0.5*math.pi
+            elif self.dy < 0: theta_radians = 1.5*math.pi
+            self.magnitude = self.dy
+        else:
+            self.magnitude = 1./(math.cos(theta_radians))*self.dx
+        theta = math.degrees(theta_radians)
+        self.heading = self._angle_or_heading(theta) # Convert angle to heading
+    
+    def almost_equals(self, other, e = 0.000001):
+        '''
+        Sometimes required for comparing GeoVectors if float error has
+        occurred. Determine if self and other (another GeoVector) are
+        equal to within e km of each other in dx and dy. The default 
+        (e = 0.000001) will return True if self and other are less than 
+        1 mm apart in distance.
+        '''
+        return abs(self.dx - other.dx) < e and abs(self.dy - other.dy) < e
         
     def __neg__(self):
         return GeoVector(-self.dx, -self.dy)
@@ -560,7 +655,7 @@ class GeoVector:
 
     def __mul__(self, other):
         # other is a scalar
-        return GeoVector(initial_heading = self.theta, distance = self.magnitude * other)
+        return GeoVector(initial_heading = self.heading, distance = self.magnitude * other)
     
     def __imul__(self, other):
         # other is a scalar
@@ -572,7 +667,7 @@ class GeoVector:
     
     def __div__(self, other):
         # other is a scalar
-        return GeoVector(initial_heading = self.theta, distance = self.magnitude / other)
+        return GeoVector(initial_heading = self.heading, distance = self.magnitude / other)
     
     def __idiv__(self, other):
         # other is a scalar
@@ -583,10 +678,10 @@ class GeoVector:
         return self.__div__(other)
     
     def __str__(self):
-        return '%s %s' %(self.theta, self.magnitude)
+        return '%s, %s' %(self.heading, self.magnitude)
     
     def __repr__(self):
-        return 'Heading %s, Distance %s' %(self.theta, self.magnitude)
+        return 'Heading %s, Distance %s' %(self.heading, self.magnitude)
     
     def type(self):
         '''
@@ -594,9 +689,11 @@ class GeoVector:
         '''
         return 'GeoVector'
 
-def test():
+def demonstration():
     palmyra = LatLon(Latitude(5.8833), Longitude(-162.0833)) # Try instantiating Latitude and Longitude objects in call
+    palmyra # Returns 'Latitude 5.8833, Longitude -162.0833'
     palmyra = LatLon(5.8833, -162.0833) # Or even simpler - initialize from two scalars expressing decimal degrees
+    print str(palmyra) # Returns '5.8833, -162.0833'
     palmyra = LatLon(Latitude(degree = 5, minute = 52, second = 59.88), Longitude(degree = -162, minute = -4.998)) # or more complicated!
     print palmyra.to_string('d% %m% %S% %H') # Print coordinates to degree minute second (returns ('5 52 59.88 N', '162 4 59.88 W'))
     palmyra = string2latlon('5 52 59.88 N', '162 4 59.88 W', 'd% %m% %S% %H') # Initialize from more complex string
@@ -614,8 +711,8 @@ def test():
     vector = (honolulu - palmyra) * 2 # A GeoVector with heading equal to the vector between palmyra and honolulu, but 2x the magnitude
     print vector # Print heading and magnitude
     print palmyra + (vector/2.0) # Recreate the coordinates of Honolulu by adding half of vector to palmyra
-    print 'Finished test!'
+    print 'Finished running demonstration!'
 
 if __name__ == '__main__':
-    test()
+    demonstration()
         
\ No newline at end of file
